"""Task module."""
from typing import Optional, Any
import time

from pydantic import BaseModel
from tqdm.auto import tqdm
import background

from shaarpec.utils import Debugger


class Task(BaseModel):
    """A running task."""

    service: str
    task_id: str
    submitted_at: str
    status: str
    success: Optional[bool]
    progress: Optional[float]
    result: Optional[Any]
    error: Optional[Any]
    debugger: Optional[Debugger]

    @background.task
    def print(self, update_interval: float = 0.1) -> None:
        """Print the progress of the task."""
        initial = 0 if self.progress is None else int(100 * self.progress)

        with tqdm(initial=initial, total=100) as pbar:
            pbar.set_description(f"Task {self.status}")

            while self.status in ("submitted", "queued", "in_progress"):
                time.sleep(update_interval)
                fraction = 0 if self.progress is None else self.progress
                pbar.update(int(100 * fraction) - pbar.n)
                pbar.set_description(f"Task {self.status}")

            pbar.set_description(
                f"Task completed ({'successfully' if self.success else 'failed'})"
            )

    def wait_for_result(self, poll_interval: float = 0.1) -> Optional[Any]:
        """Return when result becomes available."""
        while self.status in ("submitted", "queued", "in_progress"):
            time.sleep(poll_interval)

        if not self.success:
            raise ValueError(
                f"Task failed with status {self.status} and "
                f"{'error: ' + self.error if self.error else 'no message.'}"
            )

        return self.result



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['shaarpec']

package_data = \
{'': ['*']}

install_requires = \
['Babel>=2.12.1,<3.0.0',
 'anybadge>=1.14.0,<2.0.0',
 'anyio>=3.6.2,<4.0.0',
 'argon2-cffi-bindings>=21.2.0,<22.0.0',
 'argon2-cffi>=21.3.0,<22.0.0',
 'asttokens>=2.2.1,<3.0.0',
 'attrs>=22.2.0,<23.0.0',
 'backcall>=0.2.0,<0.3.0',
 'background>=0.2.1,<0.3.0',
 'beautifulsoup4>=4.12.0,<5.0.0',
 'bleach>=6.0.0,<7.0.0',
 'certifi>=2022.12.7,<2023.0.0',
 'cffi>=1.15.1,<2.0.0',
 'charset-normalizer>=3.1.0,<4.0.0',
 'click>=8.1.3,<9.0.0',
 'debugpy>=1.6.6,<2.0.0',
 'decorator>=5.1.1,<6.0.0',
 'defusedxml>=0.7.1,<0.8.0',
 'dill>=0.3.6,<0.4.0',
 'ecdsa>=0.18.0,<0.19.0',
 'entrypoints>=0.4,<0.5',
 'executing>=1.2.0,<2.0.0',
 'fastjsonschema>=2.16.3,<3.0.0',
 'httpx>=0.23.3,<0.24.0',
 'ipywidgets>=7.6.5,<8.0.0',
 'jupyterlab-widgets>=3.0.6,<4.0.0',
 'oidcish>=0.3.0,<0.4.0',
 'pylint>=2.17.1,<3.0.0',
 'tqdm>=4.65.0,<5.0.0']

setup_kwargs = {
    'name': 'shaarpec',
    'version': '2.3.0',
    'description': 'Client for SHAARPEC Analytics API.',
    'long_description': '\n<!-- PROJECT SHIELDS -->\n<!--\n*** I\'m using markdown "reference style" links for readability.\n*** Reference links are enclosed in brackets [ ] instead of parentheses ( ).\n*** See the bottom of this document for the declaration of the reference variables\n*** for contributors-url, forks-url, etc. This is an optional, concise syntax you may use.\n*** https://www.markdownguide.org/basic-syntax/#reference-style-links\n-->\n[![Contributors][contributors-shield]][contributors-url]\n[![Forks][forks-shield]][forks-url]\n[![Stargazers][stars-shield]][stars-url]\n[![Issues][issues-shield]][issues-url]\n[![MIT License][license-shield]][license-url]\n\n\n<!-- PROJECT LOGO -->\n<br />\n<div align="center">\n  <a href="https://github.com/SHAARPEC/shaarpec-python-client">\n    <img src="images/logo.png" alt="Logo" width="248" height="95">\n  </a>\n\n  <p align="center">\n    Python client for SHAARPEC Analytics API.\n    <br />\n    <a href="https://github.com/SHAARPEC/shaarpec-python-client"><strong>Explore the docs »</strong></a>\n    <br />\n    <br />\n    <a href="https://github.com/SHAARPEC/shaarpec-python-client">View Demo</a>\n    ·\n    <a href="https://github.com/SHAARPEC/shaarpec-python-client/issues">Report Bug</a>\n    ·\n    <a href="https://github.com/SHAARPEC/shaarpec-python-client/issues">Request Feature</a>\n  </p>\n</div>\n\n\n\n<!-- TABLE OF CONTENTS -->\n<details>\n  <summary>Table of Contents</summary>\n  <ol>\n    <li>\n      <a href="#about-the-project">About The Project</a>\n      <ul>\n        <li><a href="#built-with">Built With</a></li>\n      </ul>\n    </li>\n    <li>\n      <a href="#getting-started">Getting Started</a>\n      <ul>\n        <li><a href="#prerequisites">Prerequisites</a></li>\n        <li><a href="#installation">Installation</a></li>\n      </ul>\n    </li>\n    <li><a href="#usage">Usage</a></li>\n    <li><a href="#roadmap">Roadmap</a></li>\n    <li><a href="#license">License</a></li>\n    <li><a href="#contact">Contact</a></li>\n  </ol>\n</details>\n\n\n\n<!-- ABOUT THE PROJECT -->\n## About The Project\n\n[![SHAARPEC API screenshot][product-screenshot]](https://www.shaarpec.com)\n\nThis is a Python client for simple access to the SHAARPEC Analytics API. Authentication is handled automatically via device flow, authorization code flow, or client credentials flow. Authentication can also be disabled if accessing a public Analytics API.\n\nThe SHAARPEC Analytics API provides calculations on the healthcare organization\'s resources, capacities, clinical outcomes, and much more. These results can be accessed via a standard REST API, which is usually protected by the SHAARPEC Identity Server.\n\n<p align="right">(<a href="#top">back to top</a>)</p>\n\n### Built With\n\n* [![Httpx][Httpx]][Httpx]\n* [![Python][Python]][Python-url]\n\n<p align="right">(<a href="#top">back to top</a>)</p>\n\n<!-- GETTING STARTED -->\n## Getting Started\n\n### Prerequisites\n\nThe shaarpec client is used as a standard Python library. It is always a good idea to install the library in a virtual environment.\n\n### Installation\n\n1. Install the library into your virtual environment.\n   ```bash\n   pip install shaarpec\n   ```\n2. Store your credentials to the SHAARPEC IdentityServer in an .env file.\n   ```bash\n   $ cat .env\n   OIDCISH_HOST="https://idp.example.com"\n   OIDCISH_CLIENT_ID="my client id"\n   OIDCISH_CLIENT_SECRET="my client secret"\n   OIDCISH_AUDIENCE="shaarpec_api.full_access_scope"\n   OIDCISH_SCOPE="openid shaarpec_api.full_access_scope offline_access"\n   ```\n\n<p align="right">(<a href="#top">back to top</a>)</p>\n\n<!-- USAGE EXAMPLES -->\n## Usage\n\nThis library provides a Client class to easily interact with the SHAARPEC Analytics API. The class methods `Client.with_device(...)`,  `Client.with_code(...)`, and `Client.with_credentials(...)` create clients that authenticate with the SHAARPEC IdentityServer with either device flow (not tied to an individual user, recommended), code flow (tied to an individual user, for debugging and development), or credentials flow (non-interactive). There is also a class method `Client.without_auth(...)` that does not invoke the IDP server (but will only work if the Analytics API is public, otherwise give 401 Authentication invalid errors).\n\nAll API data is returned as `httpx.Response` objects.\n\nLet\'s look at some code examples on how to get data from the Analytics API. First, import the client.\n```python\nfrom shaarpec import Client\n```\n\nNext, use [device flow](https://auth0.com/docs/get-started/authentication-and-authorization-flow/device-authorization-flow) or [code flow](https://auth0.com/docs/get-started/authentication-and-authorization-flow/authorization-code-flow) to connect the client to the API with the `Client.with_device(...)`, `Client.with_code(...)`, and/or `Client.with_credentials(...)` class methods.\n\nThe credentials can either be stored in a .env file in the working directory (as explained in the Prerequisites section), provided as a path, or given directly as arguments to the `auth` dict.\n```python\n# Create a client with device flow, give authentication details directly.\nclient = Client.with_device(\n        host="https://api.shaarpec.com/",\n        auth={\n            "host": "https://idp.shaarpec.com",\n            "client_id": ...,\n            "client_secret": ...,\n            "scope": ...,\n            "audience": ...\n        }\n    )\n# Create a client with device flow, read authentication details from .env file.\nclient = Client.with_device(host="https://api.shaarpec.com/", auth="path/to/.env")\n```\nHere `host` is the base URL to the Analytics API and `auth` is a dictionary with the login credentials. With device flow, the user needs to finish the sign-in by visiting a url provided by the IdentityServer. A message will be shown:\n\n> Visit https://idp.shaarpec.com/device?userCode=XXXXXXXXX to complete sign-in.\n\nThe user visits the website, verifies that the user code is correct and confirms the sign-in. After a few seconds, the client will confirm the sign-in:\n\n> SUCCESS: Authentication was successful. Took XX.Y seconds.\n\nThe client is now connected to the API. Visit the Analytics API Base URL to interactively test the endpoints and read their documentation and about their path and query parameters. These parameters are used in the regular (`requests` and `httpx`) way with `client.verb` calls, where `verb` is either `get` or `post`. \n\n### GET and POST\n\nThe `get` and `post` verbs are supported in the standard way. For example (API responses are returned as `httpx.Response` objects):\n```python\nclient.get("terminology/allergy_type").json()\n```\nmight return\n```\n{\'419263009\': \'Allergy to tree pollen\',\n \'420174000\': \'Allergy to wheat\',\n \'425525006\': \'Allergy to dairy product\',\n \'714035009\': \'Allergy to soya\',\n \'419474003\': \'Allergy to mould\',\n \'232347008\': \'Dander (animal) allergy\',\n \'91934008\': \'Allergy to nut\',\n \'417532002\': \'Allergy to fish\',\n \'300913006\': \'Shellfish allergy\',\n \'232350006\': \'House dust mite allergy\',\n \'418689008\': \'Allergy to grass pollen\',\n \'91935009\': \'Allergy to peanuts\',\n \'91930004\': \'Allergy to eggs\',\n \'300916003\': \'Latex allergy\',\n \'424213003\': \'Allergy to bee venom\'}\n```\nor\n```python\nclient.get("population", conditions=["T78.2", "K81.0"])\n```\nmight return\n```\n[{\'patient_origin_id\': \'4c92f494-3c98-f8dd-1473-da9eb0196f6f\',\n    \'age\': \'10-16\',\n    \'is_alive\': True,\n    \'gender\': \'F\',\n    \'deceased_year\': 0},\n ...\n]\n```\n\n### Running tasks\n\nSHAARPEC Analytics API supports long-running tasks by `POST`:ing to `/service/path/to/endpoint`, and then polling with `GET` to `/service/tasks/{task_id}/status` until the result becomes available at `/service/tasks/{task_id}/results`. There is a `run` function in the library that performs this pattern.\n\nFor example\n\n```python\ntask = client.run("population/conditions")\n```\n\nwill return a task with the comorbidities in the entire population. A task is a [Pydantic](https://docs.pydantic.dev/) model with the following properties:\n\n```python\nclass Task(BaseModel):\n    """A running task."""\n    service: str\n    task_id: str\n    submitted_at: str\n    status: str\n    success: Optional[bool]\n    progress: Optional[float]\n    result: Optional[Any]\n    error: Optional[Any]\n    debugger: Optional[Any]\n```\nAs you can see, the success, progress, result and error are optional and updated automatically when available. The method comes with a progress bar which can be disabled via `client.run("path/to/task", progress_bar=False)`. If you want to use the task result in a subsequent command, you can wait (blocking) for the result with the `task.wait_for_result()` method:\n\n```python\ntask = client.run("path/to/task")\nprint(f"The result is: {task.wait_for_result()}!")\n```\n\n\n\n<p align="right">(<a href="#top">back to top</a>)</p>\n\n<!-- ROADMAP -->\n## Roadmap\n\nSee the [open issues](https://github.com/SHAARPEC/shaarpec-python-client/issues) for a full list of proposed features (and known issues).\n\n<p align="right">(<a href="#top">back to top</a>)</p>\n\n\n<!-- LICENSE -->\n## License\n\nDistributed under the MIT License. See `LICENSE` for more information.\n\n<p align="right">(<a href="#top">back to top</a>)</p>\n\n\n<!-- CONTACT -->\n## Contact\n\nSHAARPEC Support - support@shaarpec.com\n\nProject Link: [https://github.com/SHAARPEC/shaarpec-python-client](https://github.com/SHAARPEC/shaarpec-python-client)\n\n<p align="right">(<a href="#top">back to top</a>)</p>\n\n\n<!-- MARKDOWN LINKS & IMAGES -->\n<!-- https://www.markdownguide.org/basic-syntax/#reference-style-links -->\n[contributors-shield]: https://img.shields.io/github/contributors/SHAARPEC/shaarpec-python-client.svg?style=for-the-badge\n[contributors-url]: https://github.com/SHAARPEC/shaarpec-python-client/graphs/contributors\n[forks-shield]: https://img.shields.io/github/forks/SHAARPEC/shaarpec-python-client.svg?style=for-the-badge\n[forks-url]: https://github.com/SHAARPEC/shaarpec-python-client/network/members\n[stars-shield]: https://img.shields.io/github/stars/SHAARPEC/shaarpec-python-client.svg?style=for-the-badge\n[stars-url]: https://github.com/SHAARPEC/shaarpec-python-client/stargazers\n[issues-shield]: https://img.shields.io/github/issues/SHAARPEC/shaarpec-python-client.svg?style=for-the-badge\n[issues-url]: https://github.com/SHAARPEC/shaarpec-python-client/issues\n[license-shield]: https://img.shields.io/github/license/SHAARPEC/shaarpec-python-client?style=for-the-badge\n[license-url]: https://github.com/SHAARPEC/shaarpec-python-client/blob/master/LICENSE\n[product-screenshot]: images/screenshot.png\n[Httpx]: images/httpx.svg\n[Httpx-url]: https://www.python-httpx.org/\n[Python]: images/python-3.8.svg\n[Python-url]: https://www.python.org/\n',
    'author': 'Erik G. Brandt',
    'author_email': 'erik.brandt@shaarpec.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



"""Client for SHAARPEC Analytics API."""
from __future__ import annotations
from typing import Optional, Union, Any
import pathlib
import time

import background
import httpx

from oidcish import DeviceFlow, CodeFlow, CredentialsFlow

from shaarpec.tasks import Task


class Client:
    """Client for SHAARPEC Analytics API.

    The input is the base URL to the Analytics API and authentication credentials via device or
    code flow. The `.with_device(...)` and `.with_code(...)` class methods are provided and can
    be used to construct a client. See examples of usage.

    API data is returned as `httpx.Response` objects.
    \f
    Examples
    --------
    >>> from shaarpec import Client
    >>> client = Client.with_device(
            host="https://api.shaarpec.com/",
            auth={
                "host": "https://idp.shaarpec.com"
                "client_id": ...,
                "client_secret": ...,
                "audience": ...,
                "scope": ...
            }
        )
    # Or
    # >>> client = Client.with_device(
    #        host="https://api.shaarpec.com/",
    #        auth="./path/to/my/env.file"
    #    )
    #
    >>> client.get("terminology/allergy_type").json()
    {'419263009': 'Allergy to tree pollen',
     '420174000': 'Allergy to wheat',
     '425525006': 'Allergy to dairy product',
     '714035009': 'Allergy to soya',
     '419474003': 'Allergy to mould',
     '232347008': 'Dander (animal) allergy',
     '91934008': 'Allergy to nut',
     '417532002': 'Allergy to fish',
     '300913006': 'Shellfish allergy',
     '232350006': 'House dust mite allergy',
     '418689008': 'Allergy to grass pollen',
     '91935009': 'Allergy to peanuts',
     '91930004': 'Allergy to eggs',
     '300916003': 'Latex allergy',
     '424213003': 'Allergy to bee venom'}
    >>> client.get("population", conditions=["T78.2", "K81.0"])
    [{'patient_origin_id': '4c92f494-3c98-f8dd-1473-da9eb0196f6f',
        'age': '10-16',
        'is_alive': True,
        'gender': 'F',
        'deceased_year': 0},
    ...
    """

    def __init__(
        self, host: str, auth: Optional[Union[CodeFlow, DeviceFlow]], **kwargs
    ) -> None:
        timeout = kwargs.pop("timeout", 60)
        self._auth = auth
        self._client = httpx.Client(
            base_url=host,
            timeout=timeout,
            headers={"accept": "application/json"},
            **kwargs,
        )

    @classmethod
    def with_device(
        cls, host: str, auth: Union[dict[str, Any], str, None], **kwargs
    ) -> Client:
        """Authenticate with IDP host using device flow.

        The IDP host must support device flow. Authentication can be provided to
        `auth` as a dict, as environment variables, or as the path to an env file. The
        environment variables are always prefixed with OIDCISH, so OIDCISH_CLIENT_ID
        etc.
        \f
        Parameters
        ----------
          host : str
            The Analytics API base URL.
          auth : dict or string
            Authentication details and other arguments.

            If dict, then valid keywords in this dict are:
              host: str, The IDP host name (OIDCISH_HOST).
              client_id: str, The client ID (OIDCISH_CLIENT_ID).
              client_secret: str, The client secret (OIDCISH_CLIENT_SECRET).
              scope: str, A space separated, case-sensitive list of scopes
                (OIDCISH_SCOPE).
                          (default: openid profile offline_access)
              audience: str, The access claim was designated for this audience
                (OIDCISH_AUDIENCE).

            If string, then path to file with the corresponding variables.


        Examples
        --------
        >>> from shaarpec import Client
        >>> client = Client.with_device(
                "https://api.shaarpec.com",
                auth={
                    "host": "https://idp.example.com",
                    "client_id": ...,
                    "client_secret": ...,
                    "scope": ...,
                    "audience": ...,
                }
            )
        # Or, read auth variables from env file
        >>> client = Client.with_device(
                host="https://api.shaarpec.com",
                auth="path/to/my/file.env"
            )
        >>> client.get("terminology/allergy_type").json()
        ...
        """
        match auth:
            case None:
                return cls(host=host, auth=None, **kwargs)
            case str():
                return cls(host=host, auth=DeviceFlow(_env_file=auth), **kwargs)
            case dict():
                auth_host = auth.pop("host")
                return cls(host=host, auth=DeviceFlow(host=auth_host, **auth), **kwargs)
            case _:
                raise TypeError(
                    f"Object {auth} is not of recognized type ({type(auth)})."
                )

    @classmethod
    def with_code(
        cls, host: str, auth: Optional[dict[str, Any]] = None, **kwargs
    ) -> Client:
        """Authenticate with IDP server using code flow.

        The IDP host must support authorization code flow. Authentication can be
        provided to `auth` as a dict, as environment variables, or as the path to an
        env file. The environment variables are always prefixed with OIDCISH, so
        OIDCISH_CLIENT_ID etc.
        \f
        Parameters
        ----------
          host : str
             The Analytics API base URL.
          auth : dict or str,
            Authentication details and other arguments.

            If dict, then valid keyword in this dict are:
              host: str, The IDP host name (OIDCISH_HOST).
              client_id: str, The client ID (OIDCISH_CLIENT_ID).
              client_secret: str, The client secret (OIDCISH_CLIENT_SECRET).
              redirect_uri: str, Must exactly match one of the allowed redirect URIs
                for the client (OIDCISH_REDIRECT_URI).
                (default: http://localhost)
              username: str = The user name (OIDCISH_USERNAME).
              password: str = The user password (OIDCISH_PASSWORD).
              scope: str, A space separated, case-sensitive list of scopes
                (OIDCISH_SCOPE).
                          (default: openid profile offline_access)
              audience: str = The access claim was designated for this audience
                (OIDCISH_AUDIENCE).

        Examples
        --------
        >>> from shaarpec import Client
        >>> client = Client.with_code(
                "https://api.shaarpec.com",
                auth={
                    "host": "https://idp.example.com",
                    "client_id": ...,
                    "client_secret": ...,
                    "redirect_uri": ...,
                    "username": ...,
                    "password": ...,
                    "scope": ...,
                    "audience": ...,
                }
            )
        # Or, read auth variables from env file
        >>> client = Client.with_code(
                host="https://api.shaarpec.com",
                auth="path/to/my/file.env"
            )
        >>> client.get("terminology/allergy_type").json()
        ...
        """
        match auth:
            case None:
                return cls(host=host, auth=None, **kwargs)
            case str():
                return cls(host=host, auth=CodeFlow(_env_file=auth), **kwargs)
            case dict():
                auth_host = auth.pop("host")
                return cls(host=host, auth=CodeFlow(host=auth_host, **auth), **kwargs)
            case _:
                raise TypeError(
                    f"Object {auth} is not of recognized type ({type(auth)})."
                )

    @classmethod
    def with_credentials(
        cls, host: str, auth: Union[dict[str, Any], str, None], **kwargs
    ) -> Client:
        """Authenticate with IDP host using client credentials flow.

        The IDP host must support client credentials flow. Authentication can be provided to
        `auth` as a dict, as environment variables, or as the path to an env file. The
        environment variables are always prefixed with OIDCISH, so OIDCISH_CLIENT_ID
        etc.
        \f
        Parameters
        ----------
          host : str
            The Analytics API base URL.
          auth : dict or string
            Authentication details and other arguments.

            If dict, then valid keywords in this dict are:
              host: str, The IDP host name (OIDCISH_HOST).
              client_id: str, The client ID (OIDCISH_CLIENT_ID).
              client_secret: str, The client secret (OIDCISH_CLIENT_SECRET).
              audience: str, The access claim was designated for this audience
                (OIDCISH_AUDIENCE).

            If string, then path to file with the corresponding variables.

        Examples
        --------
        >>> from shaarpec import Client
        >>> client = Client.with_credentials(
                "https://api.shaarpec.com",
                auth={
                    "host": "https://idp.example.com",
                    "client_id": ...,
                    "client_secret": ...,
                    "audience": ...,
                }
            )
        # Or, read auth variables from env file
        >>> client = Client.with_credentials(
                host="https://api.shaarpec.com",
                auth="path/to/my/file.env"
            )
        >>> client.get("terminology/allergy_type").json()
        ...
        """
        match auth:
            case None:
                return cls(host=host, auth=None, **kwargs)
            case str():
                return cls(host=host, auth=CredentialsFlow(_env_file=auth), **kwargs)
            case dict():
                auth_host = auth.pop("host")
                return cls(host=host, auth=CredentialsFlow(host=auth_host, **auth), **kwargs)
            case _:
                raise TypeError(
                    f"Object {auth} is not of recognized type ({type(auth)})."
                )

    @classmethod
    def without_auth(cls, host: str, **kwargs) -> Client:
        """Create a client that does not use authentication.

        \f
        Parameters
        ----------
          host : str
            The Analytics API base URL.

        Examples
        --------
        >>> from shaarpec import Client
        >>> client = Client.without_auth("https://api.shaarpec.com")
        >>> client.get("terminology/allergy_type").json()
        ...
        """
        return cls(host=host, auth=None, **kwargs)

    @property
    def auth(self) -> Optional[Union[CodeFlow, DeviceFlow]]:
        """Return the authentication credentials."""
        return self._auth

    def get(self, uri: str, **kwargs) -> httpx.Response:
        """Get the resource at `uri`.

        Keyword arguments are passed as query parameters.
        """
        headers = {}

        if (auth := self.auth) is not None:
            if (credentials := auth.credentials) is not None:
                access_token = credentials.access_token

                headers = {
                    "Authorization": f"Bearer {access_token}",
                    "x-auth-request-access-token": access_token,
                }

        response = self._client.get(
            uri,
            headers=headers,
            params=kwargs,
        )

        return response

    # pylint: disable=too-many-arguments
    def post(
        self, uri: str, content=None, data=None, files=None, json=None, **kwargs
    ) -> httpx.Response:
        """Post to the resource at `uri`.

        Keyword arguments are passed as query parameters.
        """
        headers = {}

        if (auth := self.auth) is not None:
            if (credentials := auth.credentials) is not None:
                access_token = credentials.access_token

                headers = {
                    "Authorization": f"Bearer {access_token}",
                    "x-auth-request-access-token": access_token,
                }

        response = self._client.post(
            uri,
            content=content,
            data=data,
            files=files,
            json=json,
            headers=headers,
            params=kwargs,
        )

        return response

    def run(
        self,
        uri: str,
        content=None,
        data=None,
        files=None,
        json=None,
        poll_interval: float = 0.1,
        progress_bar: bool = True,
        **kwargs,
    ) -> Task:
        """Run a task at resource `uri`.

        A task is equivalent to posting and polling until the job is finished.

        Keyword arguments are passed as query parameters.
        """
        service = pathlib.Path(uri).parts[0]

        response = self.post(
            uri=uri, content=content, data=data, files=files, json=json, **kwargs
        )

        match response.status_code:
            case 202:
                pass
            case 401:
                raise RuntimeError("401: Not authenticated.")
            case _:
                raise RuntimeError(
                    "Did not receive 202 Accepted for the task. "
                    f"The actual response code was {response.status_code}. "
                    f"The actual message was {response.text}."
                )

        task_id = response.json().get("task_id")
        submitted_at = response.json().get("submitted_at")

        task = Task(
            service=service,
            task_id=task_id,
            submitted_at=submitted_at,
            status="submitted",
            success=None,
            progress=None,
            result=None,
            error=None,
            debugger=None,
        )

        if progress_bar:
            task.print()

        self._wait_for_task(task, poll_interval=poll_interval)

        return task

    @background.task
    def _wait_for_task(self, task: Task, poll_interval: float = 0.1) -> None:
        while task.status in ("submitted", "queued", "in_progress"):
            response = self.get(f"{task.service}/tasks/{task.task_id}/status")

            match response.status_code:
                case 401:
                    print(
                        f"Not authorized to run task id {task.task_id} on service {task.service}."
                    )
                    task.success = False
                    task.status = "unauthorized"

                case 404:
                    print(
                        f"Task with id {task.task_id} could not be found on service {task.service}."
                    )
                    task.success = False
                    task.status = "not_found"

                case 200:
                    match response.json():
                        case {"status": "not_found"}:
                            print(
                                f"Task with id {task.task_id} "
                                f"could not be found on service {task.service}."
                            )
                            task.status = "not_found"

                        case {"status": "in_progress" | "queued"} as data:
                            task.progress = data.get("progress")
                            task.status = data.get("status")
                            time.sleep(poll_interval)

                        case {"status": "complete", "success": True} as data:
                            resp = self.get(
                                f"{task.service}/tasks/{task.task_id}/results"
                            )

                            task.result = resp.json().get("result")
                            task.progress = 1.0
                            task.success = True
                            task.status = "complete"

                        case _ as data:
                            task.error = data.get("error")
                            task.success = False
                            task.status = "complete"

                case _:
                    print(f"An error occurred on service {task.service}.")
                    task.error = response.json().get("error")
                    task.success = False
                    task.status = "error"
